# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import errno
import struct
import sys
import typing as t

from .commons import TrackSector, hex_dump
from .native import NativeFile

if t.TYPE_CHECKING:
    from .abstract import AbstractFile

___all__ = [
    "IMDDiskImage",
    "IMDTrack",
    "IMDSector",
]

# ImageDisk, A Diskette Imaging program for Soft-Sector Formats, Pag 26
# https://oldcomputers-ddns.org/public/pub/manuals/imd.pdf

EOF = 0x1A  # End of file marker in IMD files
TRACK_HEADER_FORMAT = "BBBBB"
TRACK_HEADER_SIZE = struct.calcsize(TRACK_HEADER_FORMAT)

SECTOR_HEAD_MAP = 0x40  # Bit indicating presence of sector head map
SECTOR_CYLINDER_MAP = 0x80  # Bit indicating presence of sector cylinder map

# Sector Data Record Type
UNAVAILABLE = 0  # Sector data unavailable - could not be read
NORMAL = 1  # Normal data: (Sector Size) bytes follow
COMPRESSED = 2  # Compressed: All bytes in sector have same value (xx)
DELETED = 3  # Normal data with "Deleted-Data address mark"
DELETED_COMPRESSED = 4  # Compressed with "Deleted-Data address mark"
DATA_ERROR = 5  # Normal data read with data error
DATA_ERROR_COMPRESSED = 6  # Compressed read with data error
DATA_ERROR_DELETED = 7  # Deleted data read with data error
DATA_ERROR_DELETED_COMPRESSED = 8  # Compressed, Deleted read with data error

RECORD_TYPES = {
    UNAVAILABLE: "Unavailable",
    NORMAL: "Normal",
    COMPRESSED: "Compressed",
    DELETED: "Deleted",
    DELETED_COMPRESSED: "Deleted, Compressed",
    DATA_ERROR: "Data Error",
    DATA_ERROR_COMPRESSED: "Data Error, Compressed",
    DATA_ERROR_DELETED: "Data Error, Deleted",
    DATA_ERROR_DELETED_COMPRESSED: "Data Error, Deleted, Compressed",
}


class IMDSector:
    """
    Sector Data Record

    +---------------------------------------+
    | Sector Data Record Type               |  1 byte
    +---------------------------------------+
    | Data                                  |  record_size - 1 bytes
    /                                       /
    |                                       |
    +---------------------------------------+

    """

    track: "IMDTrack"  # Reference to the track this sector belongs to
    record_type: int  # Type of the record
    record_size: int  # Record size in bytes, including the record type byte
    data: bytes  # The actual data of the sector

    def __init__(self, track: "IMDTrack") -> None:
        self.track = track
        self.record_type = 0
        self.record_size = 0
        self.data = b""

    @classmethod
    def read(cls, track: "IMDTrack", buffer: bytes, position: int) -> "IMDSector":
        self = cls(track)
        self.record_type = int(buffer[position])
        # Determine the size of the record based on its type
        if self.record_type == UNAVAILABLE:
            # 1 byte for the record type, no data
            self.record_size = 1 + 0
            self.data = b""
        elif self.record_type in (
            COMPRESSED,
            DELETED_COMPRESSED,
            DATA_ERROR_COMPRESSED,
            DATA_ERROR_DELETED_COMPRESSED,
        ):
            # 1 byte for the record type, 1 byte for compressed data
            self.record_size = 1 + 1
            tmp = buffer[position + 1]
            self.data = bytes([tmp] * track.sector_size)  # Fill with the compressed value
        else:
            # 1 byte for the record type, sector_size bytes for data
            self.record_size = 1 + track.sector_size
            self.data = buffer[position + 1 : position + self.record_size]
        return self

    def __str__(self) -> str:
        return f"{self.record_size:5} bytes {RECORD_TYPES[self.record_type]}"


class IMDTrack:
    """
    Track
    ImageDisk, A Diskette Imaging program for Soft-Sector Formats, Pag 26
    https://oldcomputers-ddns.org/public/pub/manuals/imd.pdf

    +---------------------------------------+
    |  Mode value (0-5)                     |   1 byte
    +---------------------------------------+
    |  Cylinder (0-n)                       |   1 byte
    +---------------------------------------+
    |  SCM  |  SHM  | Head (0-1)            |   1 byte
    +---------------------------------------+
    |  Number of sectors in track (1-n)     |   1 byte
    +---------------------------------------+
    |  Sector size (0-6)                    |   1 byte
    +---------------------------------------+
    |  Sector numbering map                 |   sectors_per_track bytes
    /                                       /
    +---------------------------------------+
    |  Sector cylinder map (optional)       |   sectors_per_track bytes (if present)
    /                                       /
    +---------------------------------------+
    |  Sector head map (optional)           |   sectors_per_track bytes (if present)
    /                                       /
    +---------------------------------------+
    |  Sector data records                  |   variable length
    /                                       /
    +---------------------------------------+

    """

    disk: "IMDDiskImage"  # Reference to the disk image
    mode: int  # Data transfer rate and density
    cylinder: int  # Cylinder number
    head: int  # Head number (0 or 1)
    sectors_per_track: int  # Number of sectors in this track
    sector_size: int  # Size of each sector in bytes
    sector_numbering_map: t.List[int]  # Map of sector numbers
    sector_cylinder_map: t.Optional[t.List[int]]  # Sector cylinder map
    sector_head_map: t.Optional[t.List[int]]  # Sector head map
    sector_data_records: t.List["IMDSector"]  # Sector data records
    track_size: int  # Size of the track in bytes

    def __init__(self, disk: "IMDDiskImage") -> None:
        self.disk = disk
        self.mode = 0
        self.cylinder = 0
        self.head = 0
        self.sectors_per_track = 0
        self.sector_size = 0
        self.sector_numbering_map = []
        self.sector_cylinder_map = None
        self.sector_head_map = None
        self.sector_data_records = []

    @classmethod
    def read(cls, disk: "IMDDiskImage", buffer: bytes, position: int) -> "IMDTrack":
        initial_position = position
        self = cls(disk)
        (
            self.mode,
            self.cylinder,
            raw_head,
            self.sectors_per_track,
            raw_sector_size,
        ) = struct.unpack_from(TRACK_HEADER_FORMAT, buffer, position)
        self.head = raw_head & 0x01
        self.sector_size = 1 << (7 + raw_sector_size)
        position += TRACK_HEADER_SIZE
        # Read sector numbering map
        self.sector_numbering_map = list(buffer[position : position + self.sectors_per_track])
        position += self.sectors_per_track
        # Read optional sector cylinder map
        if raw_head & SECTOR_CYLINDER_MAP:
            self.sector_cylinder_map = list(buffer[position : position + self.sectors_per_track])
            position += self.sectors_per_track
        # Read optional sector head map
        if raw_head & SECTOR_HEAD_MAP:
            self.sector_head_map = list(buffer[position : position + self.sectors_per_track])
            position += self.sectors_per_track
        # Read sector data records
        for _ in range(self.sectors_per_track):
            sector_record = IMDSector.read(self, buffer, position)
            position += sector_record.record_size
            self.sector_data_records.append(sector_record)
        # Calculate the size of the track
        self.track_size = position - initial_position
        return self


class IMDDiskImage:
    """
    Represents an IMD disk image

    ImageDisk, A Diskette Imaging program for Soft-Sector Formats, Pag 26
    https://oldcomputers-ddns.org/public/pub/manuals/imd.pdf

    The overall layout of an ImageDisk .IMD image file is:

    +---------------------------------------+
    | IMD v.vv: dd/mm/yyyy hh:mm:ss         |  ASCII Header
    | Comment (ASCII only)                  |  variable length
    +---------------------------------------+
    | EOF (0x1A) byte                       |  1 byte
    +---------------------------------------+
    | Tracks Data                           |
    | ...                                   |
    /                                       /
    |                                       |
    +---------------------------------------+

    """

    f: "AbstractFile"  # File object representing the disk image
    ascii_header: str  # ASCII header of the disk image
    tracks: list["IMDTrack"]  # List of tracks in the disk image

    def __init__(self, file: "AbstractFile") -> None:
        self.f = file
        self.ascii_header = ""
        self.tracks = []

        buffer = self.f.read()
        try:
            ascii_header, track_data = buffer.split(bytes([EOF]), 1)
        except Exception:
            raise OSError(errno.EIO, "Invalid IMD file format: Missing EOF marker")
        if not ascii_header.startswith(b"IMD "):
            raise OSError(errno.EIO, "Invalid IMD file format: Missing 'IMD' header")
        self.ascii_header = ascii_header.decode("ascii")
        # Read the tracks data
        position = 0
        while position < len(track_data):
            track = IMDTrack.read(self, track_data, position)
            position += track.track_size
            self.tracks.append(track)

    def read_sector(self, address: TrackSector) -> bytes:
        """
        Read a sector from the disk image
        """
        track = self.tracks[address.track]
        sector_index = track.sector_numbering_map[address.sector]
        sector = track.sector_data_records[sector_index]
        return sector.data


def main() -> None:
    if len(sys.argv) < 2:
        print("Usage: python imd.py <path_to_imd_file> [track sector]")
        sys.exit(1)
    file = NativeFile(sys.argv[1])
    disk = IMDDiskImage(file)
    if len(sys.argv) >= 4:
        track_number = int(sys.argv[2])
        sector_number = int(sys.argv[3])
        address = TrackSector(track=track_number, sector=sector_number)
        try:
            data = disk.read_sector(address)
            print(f"Data from track {track_number}, sector {sector_number}")
            hex_dump(data)
        except IndexError:
            print(f"Invalid track {track_number} or sector {sector_number}.")
    else:
        print(f"Loaded IMD disk image with {len(disk.tracks)} tracks.")
        print(f"{disk.ascii_header}")
        for i, track in enumerate(disk.tracks):
            print(
                f"Track {i}: Mode={track.mode}, Cylinder={track.cylinder}, Head={track.head}, "
                f"Sectors={track.sectors_per_track}, Sector Size={track.sector_size} bytes"
            )
            for j, sector in enumerate(track.sector_data_records):
                print(f"  Sector {j:3}:  {sector}")



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import typing as t

__all__ = [
    "read_baudot_string",
    "str_to_baudot",
    "fiodec_to_str",
    "str_to_fiodec",
    "LABEL_END_WORD",
    "BAUDOT_TO_ASCII",
]

LABEL_END_WORD = 0o777777  # End of Baudot string


def read_baudot_string(words: t.List[int], position: int = 0) -> t.Tuple[str, int]:
    """
    Read a Baudot string, return the string and the new position.
    The strings is padded with 0 to an 18b boundary and terminated
    by a word of all ones (LABEL_END_WORD).
    """
    data = []
    while position < len(words) and words[position] != LABEL_END_WORD:
        word = words[position]
        position += 1
        for word in [
            (word >> 12) & 0o077,
            (word >> 6) & 0o077,
            (word & 0o077),
        ]:
            if word != 0:
                data.append(BAUDOT_TO_ASCII.get(word, '-'))
    return "".join(data), position


def str_to_baudot(val: str, length: t.Optional[int] = None) -> t.List[int]:
    """
    Convert a string to a Baudot word list.
    """
    data = []
    for i in range(0, len(val), 3):
        word = 0
        for j in range(0, 3):
            try:
                ch = val[i + j]
            except IndexError:
                ch = "\0"
            word = (word << 6) | (ASCII_TO_BAUDOT.get(ch, 0) & 0o77)
        data.append(word)
    if length is not None:
        # Truncate or pad the data
        data = data[:length]
        if len(data) < length:
            # Add padding
            data.extend([0] * (length - len(data)))
    return data


# https://bestengineeringprojects.com/alphanumeric-codes-description-and-types/

BAUDOT_TO_ASCII = {
    # Letters
    0b110000: 'A',
    0b011100: 'C',
    0b100110: 'B',
    0b100100: 'D',
    0b100000: 'E',
    0b101100: 'F',
    0b010110: 'G',
    0b001010: 'H',
    0b011000: 'I',
    0b110100: 'J',
    0b111100: 'K',
    0b010010: 'L',
    0b001110: 'M',
    0b001100: 'N',
    0b000110: 'O',
    0b011010: 'P',
    0b111010: 'Q',
    0b010100: 'R',
    0b101000: 'S',
    0b000010: 'T',
    0b111000: 'U',
    0b011110: 'V',
    0b110010: 'W',
    0b101110: 'X',
    0b101010: 'Y',
    0b100010: 'Z',
    0b111110: '-',  #  Figures shift
    0b110110: '-',  #  Letters shift
    0b001000: ' ',  #  Space
    0b010000: '\n',  # New line (line feed)
    0b000000: '\0',  # Null
    # Figures
    0b110001: '-',
    0b011101: '?',
    0b100111: ':',
    0b100101: '$',
    0b100001: '3',
    0b101101: '!',
    0b010111: '&',
    0b001011: '#',
    0b011001: '8',
    0b110101: '`',
    0b111101: '(',
    0b010011: ')',
    0b001111: '.',
    0b001101: '\'',
    0b000111: '9',
    0b011011: '0',
    0b111011: '1',
    0b010101: '4',
    0b101001: '\b',  # Bell
    0b000011: '5',
    0b111001: '7',
    0b011111: ';',
    0b110011: '2',
    0b101111: '/',
    0b101011: '6',
    0b100011: '"',
    0b111111: '-',  #  Figures shift
    0b110111: '-',  #  Letters shift
    0b001001: ' ',  #  Space
    0b010001: '\n',  # New line (line feed)
    0b000001: '\0',  # Null
}

ASCII_TO_BAUDOT = {v: k for k, v in BAUDOT_TO_ASCII.items() if k not in [0b001001, 0b010001, 0b000001]}


def fiodec_to_str(words: t.List[int], position: int = 0) -> str:
    """
    Convert a list of FIODEC words to a string
    """
    data: t.List[str] = []
    shift = 0
    eof = False
    for word in words[position:]:
        position += 1
        chars = [
            (word >> 12) & 0o077,
            (word >> 6) & 0o077,
            (word & 0o077),
        ]
        for i, ch in enumerate(chars):
            if i == 0:
                if ch == FIODEC_END_OF_LINE:  # End of line; next two characters are line number
                    data.append('\n')
                    break
                elif ch == FIODEC_END_OF_PAGE:  # End of page; next two characters are page number
                    data.append('\f')
                    break
                elif ch == FIODEC_END_OF_FILE:  # End of file
                    eof = True
                    break
                elif ch == FIODEC_MASTER_SPACE:
                    break
            if ch == FIODEC_MASTER_SPACE:
                pass
            elif ch == FIODEC_SHIFT_ON:
                shift = 0o100
            elif ch == FIODEC_SHIFT_OFF:
                shift = 0
            else:
                if ch + shift in FIODEC_TO_ASCII:
                    data.append(FIODEC_TO_ASCII[ch + shift])
        if eof:
            break
    return "".join(data)


def str_to_fiodec(val: str) -> t.List[int]:
    """
    Convert a string to a list of FIODEC words
    """
    data: t.List[int] = []
    shift: bool = False
    current_word: t.List[int] = []
    line_number = 1  # Line number
    page_number = 1  # Page number

    def add_current_word_to_data() -> None:
        # Add the current word to the data
        # and clear the current word
        data.append(current_word[0] << 12 | current_word[1] << 6 | current_word[2])
        current_word.clear()

    def add_ch(*args: int, flush: bool = False) -> None:
        # Add one or more characters to the current word
        if flush:
            flush_current_word()
        for v in args:
            # Add a character to the current word
            current_word.append(v & 0o77)
            # If the current word is complete, add it to the data
            if len(current_word) == 3:
                add_current_word_to_data()

    def add_end_of_line() -> None:
        # Add end of line
        add_ch(FIODEC_END_OF_LINE, line_number >> 6, line_number & 0o77, flush=True)

    def add_end_of_page() -> None:
        # Add end of page
        add_ch(FIODEC_END_OF_PAGE, page_number >> 6, page_number & 0o77, flush=True)

    def add_end_of_file() -> None:
        # Add end of file
        add_ch(FIODEC_END_OF_FILE, 0, 0, flush=True)

    def flush_current_word() -> None:
        # Flush the current word
        if current_word:
            # Pad the current word with spaces
            if len(current_word) < 3:
                current_word.extend([FIODEC_MASTER_SPACE] * (3 - len(current_word)))
            # Add the current word to the data
            add_current_word_to_data()

    for ch in val:
        if ch == '\n':  # End of line, line_number
            add_end_of_line()
            line_number += 1
            if line_number > FIODEC_LINES_PER_PAGE:
                # Add a new page
                add_end_of_page()
                line_number = 1
                page_number += 1
        elif ch == '\f':  # End of page, page_number
            add_end_of_page()
            line_number = 1
            page_number += 1
        elif ch == '\x1A':  # End of file
            break
        else:
            ch = ch.upper()
            v: int = ASCII_TO_FIODEC.get(ch, None)  # type: ignore
            if v is not None:
                if v & 0o100:
                    if not shift:
                        shift = True
                        add_ch(FIODEC_SHIFT_ON)
                    add_ch(v & 0o77)
                else:
                    if shift:
                        shift = False
                        add_ch(FIODEC_SHIFT_OFF)
                    add_ch(v)

    # Add end of page
    if line_number > 1:
        add_end_of_page()
    # Add end of file
    add_end_of_file()
    # Add one 0 word
    add_ch(0, 0, 0)
    flush_current_word()
    return data


# https://simh.trailing-edge.com/docs/decsys.pdf

FIODEC_END_OF_LINE = 0o14
FIODEC_END_OF_PAGE = 0o15
FIODEC_END_OF_FILE = 0o16
FIODEC_MASTER_SPACE = 0o17
FIODEC_SHIFT_ON = 0o74
FIODEC_SHIFT_OFF = 0o72
FIODEC_LINES_PER_PAGE = 60

FIODEC_TO_ASCII = {
    0o00: ' ',  # Space
    0o01: '1',
    0o02: '2',
    0o03: '3',
    0o04: '4',
    0o05: '5',
    0o06: '6',
    0o07: '7',
    0o10: '8',
    0o11: '9',
    0o13: '\f',  # Form feed
    0o20: '0',
    0o21: '/',
    0o22: 'S',
    0o23: 'T',
    0o24: 'U',
    0o25: 'V',
    0o26: 'W',
    0o27: 'X',
    0o30: 'Y',
    0o31: 'Z',
    0o33: ',',
    0o34: ':',
    0o36: '\t',  # Tab
    0o40: '@',
    0o41: 'J',
    0o42: 'K',
    0o43: 'L',
    0o44: 'M',
    0o45: 'N',
    0o46: 'O',
    0o47: 'P',
    0o50: 'Q',
    0o51: 'R',
    0o54: '-',
    0o55: ')',
    0o56: '\\',
    0o57: '(',
    0o61: 'A',
    0o62: 'B',
    0o63: 'C',
    0o64: 'D',
    0o65: 'E',
    0o66: 'F',
    0o67: 'G',
    0o70: 'H',
    0o71: 'I',
    0o73: '.',
    # Shift
    0o100: ' ',
    0o101: '"',
    0o102: '\'',
    0o103: '~',
    0o104: '#',
    0o105: '!',
    0o106: '&',
    0o107: '<',
    0o110: '>',
    0o111: '^',
    0o120: '`',
    0o121: '?',
    0o133: '=',
    0o134: ';',
    0o140: '_',
    0o154: '+',
    0o155: ']',
    0o156: '|',
    0o157: '[',
    0o173: '*',
}

ASCII_TO_FIODEC = {v: k for k, v in FIODEC_TO_ASCII.items() if k not in [0o100]}



import shlex

import pytest

from xferx.commons import bytes_to_word, word_to_bytes
from xferx.pdp11.rad50 import asc2rad, rad2asc
from xferx.shell.commons import PartialMatching, extract_options


def test_bytes_to_word():
    # Test with valid input
    assert bytes_to_word(b"\x01\x00") == 1
    assert bytes_to_word(b"\xff\xff") == 65535
    assert bytes_to_word(b"\x01\xab\xcd", position=1) == 52651
    # Test with out of bounds position
    with pytest.raises(IndexError):
        bytes_to_word(b"\x01\x02", position=2)


def test_word_to_bytes():
    # Test with valid input
    assert word_to_bytes(1) == b"\x01\x00"
    assert word_to_bytes(65535) == b"\xFF\xFF"
    assert len(word_to_bytes(1234)) == 2
    for i in range(0, 1 << 16):
        assert bytes_to_word(word_to_bytes(i)) == i
    # Test with negative value
    with pytest.raises(ValueError):
        word_to_bytes(-1)
    # Test with value exceeding 16-bit range
    with pytest.raises(ValueError):
        word_to_bytes(2**16)


def test_rad2asc():
    # Test with valid input
    assert rad2asc(b"\x01\x00") == "A"
    assert rad2asc(b"\x06\x01") == "FV"
    assert rad2asc(b"\x00\x00") == ""
    # Test with different positions
    assert rad2asc(b"\x10\x37\x31\x43\x74", position=0) == "H2P"
    assert rad2asc(b"\x10\x37\x31\x43\x74", position=2) == "J0A"
    # Test with all zeros
    assert rad2asc(b"\x00\x00\x00") == ""


def test_asc2rad():
    # Test with valid input
    assert asc2rad("ABC") == b"\x93\x06"
    assert asc2rad("Z12") == b"x\xa7"
    assert asc2rad("") == b"\x00\x00"
    # Test with lowercase characters
    assert asc2rad("zia") == b"\xe9\xa3"
    assert asc2rad(":$%") == b"\x54\xfe"


def test_partial_matching():
    x = PartialMatching()
    x.add("APP_LE")
    x.add("PE_AR")
    x.add("O_RANGE")
    x.add("D")
    x.add("DA_TE")

    # Test partial matching keys
    assert x.get("APP") == "APPLE"
    assert x.get("PE") == "PEAR"
    assert x.get("PEA") == "PEAR"
    assert x.get("O") == "ORANGE"
    assert x.get("OR") == "ORANGE"
    assert x.get("ORA") == "ORANGE"
    assert x.get("ORAN") == "ORANGE"
    assert x.get("D") == "D"
    assert x.get("DA") == "DATE"
    assert x.get("DAT") == "DATE"
    assert x.get("DATE") == "DATE"

    # Test non-partial matching keys
    assert x.get("XXX") is None
    assert x.get("XX") is None
    assert x.get("A") is None
    assert x.get("P") is None
    assert x.get("PE_X") is None
    assert x.get("O_") is None
    assert x.get("ORANGO") is None
    assert x.get("TD") is None
    assert x.get("") is None


def test_extract_options():
    line = "command /a /b /c:1 /d:abc /flag value1 value2"
    options = ("/a", "/b", "/c", "/d", "/flag")

    args, opts = extract_options(shlex.split(line), *options)
    assert args == ["command", "value1", "value2"]
    assert opts == {"a": True, "b": True, "c": "1", "d": "abc", "flag": True}


def test_extract_options_with_no_options():
    line = "command value1 value2"
    options = ("/a", "/b", "/flag")

    args, opts = extract_options(shlex.split(line), *options)
    assert args == ["command", "value1", "value2"]
    assert opts == {}


def test_extract_options_with_some_options():
    line = "command /a value1 /flag value2"
    options = ("/a", "/b", "/flag")

    args, opts = extract_options(shlex.split(line), *options)
    assert args == ["command", "value1", "value2"]
    assert opts == {"a": True, "flag": True}


def test_extract_options_case_insensitive():
    line = "command /A /B /FLAG value1 value2"
    options = ("/a", "/b", "/flag")

    args, opts = extract_options(shlex.split(line), *options)
    assert args == ["command", "value1", "value2"]
    assert opts == {"a": True, "b": True, "flag": True}


def test_extract_options_with_unexpected_options():
    line = "command /x /y value1 value2"
    options = ("/a", "/b", "/flag")

    args, opts = extract_options(shlex.split(line), *options)
    assert args == ["command", "/x", "/y", "value1", "value2"]
    assert opts == {}



import random
from datetime import datetime, timedelta

import pytest

from xferx.apple2.commons import (
    ProDOSFileInfo,
    decode_apple_single,
    encode_apple_single,
)
from xferx.apple2.prodosfs import (
    ProDOSFilesystem,
    date_to_prodos,
    parse_file_aux_type,
    prodos_to_date,
)
from xferx.shell import Shell

DSK = "tests/dsk/prodos.dsk"


def test_date_round_trip():
    start_year = 1980
    end_year = 2020
    for i in range(0, 100):
        start_date = datetime(start_year, 1, 1)
        end_date = datetime(end_year, 12, 31, 23, 59, 59)
        random_seconds = random.randint(0, int((end_date - start_date).total_seconds()))
        random_seconds = random_seconds - (random_seconds % 60)
        dt = start_date + timedelta(seconds=random_seconds)
        assert prodos_to_date(date_to_prodos(dt)) == dt


def test_bitmap():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /prodos {DSK}", batch=True)
    fs = shell.volumes.get('T')
    bitmap = fs.read_bitmap()
    print(f"used: {bitmap.used()} free: {bitmap.free()} total: {fs.total_blocks}")
    u = set()
    for b in fs.root.iterdir():
        if hasattr(b, "blocks"):
            u.update(b.blocks())
    for b in fs.root.iterdir():
        if hasattr(b, "blocks"):
            print(b, list(b.blocks()))
            for x in b.blocks():
                assert not bitmap.is_free(x)
    for i in range(0, fs.total_blocks):
        if bitmap.is_free(i):
            assert i not in u


def test_prodos():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /prodos {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, ProDOSFilesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir/brief t:", batch=True)
    shell.onecmd("dir/brief t:/pr", batch=True)
    with pytest.raises(FileNotFoundError):
        shell.onecmd("dir/brief t:/notfound", batch=True)
    with pytest.raises(FileNotFoundError):
        shell.onecmd("dir/brief t:/pr/notfound/test", batch=True)
    shell.onecmd("type t:/pr/small/1.txt", batch=True)

    x = fs.read_bytes("/pr/small/medium/50.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("small/medium/*.TXT"))
    assert len(l) == 3


def test_prodos_init():
    # Init
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /prodos {DSK}", batch=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"init /prodos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /prodos {DSK}.mo", batch=True)
    shell.onecmd("ou:", batch=True)
    shell.onecmd("dir", batch=True)
    shell.onecmd("create/directory aaa", batch=True)
    shell.onecmd("create/directory aaa/bbb", batch=True)
    shell.onecmd("copy t:small/medium/*.txt aaa/bbb", batch=True)
    shell.onecmd("copy t:small/1.txt ou:", batch=True)
    shell.onecmd("cd /prodos/aaa/bbb", batch=True)
    shell.onecmd("dir", batch=True)
    shell.onecmd("cd /prodos", batch=True)
    fs = shell.volumes.get('OU')

    x1 = fs.read_bytes("aaa/bbb/50.txt")
    x1 = x1.rstrip(b"\0")
    assert len(x1) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x1

    shell.onecmd("delete ou:aaa", batch=True)
    assert fs.read_bytes("1.txt")
    with pytest.raises(FileNotFoundError):
        x1 = fs.read_bytes("aaa/bbb/50.txt")
    shell.onecmd("ou:", batch=True)
    with pytest.raises(Exception):
        shell.onecmd("cd aaa/bbb", batch=True)

    # Test init mounted volume
    shell.onecmd("init ou:", batch=True)
    with pytest.raises(Exception):
        fs.read_bytes("1.txt")


def test_grow_dir():
    # Test grow directory
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount t: /prodos {DSK}", batch=True)
    shell.onecmd(f"mount ou: /prodos {DSK}.mo", batch=True)
    for i in range(0, 50):
        shell.onecmd(f"create/file ou:{i:05d}.txt /type:txt /allocate:1", batch=True)
    shell.onecmd("dir ou:", batch=True)


def test_types():
    assert parse_file_aux_type("txt,3000") == (0x4, 3000)
    assert parse_file_aux_type("bin,$2000") == (0x6, 0x2000)
    assert parse_file_aux_type("txt") == (0x4, 0)
    assert parse_file_aux_type("$99") == (0x99, 0)
    assert parse_file_aux_type("$99,$ff") == (0x99, 0xFF)

    shell = Shell(verbose=True)
    shell.onecmd(f"create {DSK}.mo /allocate:2000", batch=True)
    shell.onecmd(f"init /prodos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /prodos {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')
    shell.onecmd("create ou:test1 /allocate:1 /type:txt,3000", batch=True)
    test1 = fs.get_file_entry("test1")
    assert test1.storage_type == 0x1
    assert test1.prodos_file_type == 0x4
    assert test1.aux_type == 3000
    shell.onecmd("create ou:test2 /allocate:1 /type:bin,$2000", batch=True)
    test2 = fs.get_file_entry("test2")
    assert test2.storage_type == 0x1
    assert test2.prodos_file_type == 0x6
    assert test2.aux_type == 0x2000


def test_pascal_area():
    shell = Shell(verbose=True)
    shell.onecmd(f"create {DSK}.mo /allocate:2000", batch=True)
    shell.onecmd(f"init /prodos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /prodos {DSK}.mo", batch=True)
    shell.onecmd("create ou:pascal.area /allocate:500", batch=True)
    shell.onecmd("create ou:pascal.area/vol1 /allocate:200", batch=True)
    shell.onecmd("create ou:pascal.area/vol2 /allocate:100", batch=True)
    shell.onecmd("create ou:pascal.area/vol3 /allocate:100", batch=True)
    with pytest.raises(OSError):
        shell.onecmd("create ou:pascal.area/vol4 /allocate:1000", batch=True)
    shell.onecmd("del ou:pascal.area/vol2", batch=True)
    with pytest.raises(OSError):
        shell.onecmd("create ou:pascal.area/vol2 /allocate:101", batch=True)
    shell.onecmd("create ou:pascal.area/vol2a /allocate:50", batch=True)
    shell.onecmd("create ou:pascal.area/vol2b /allocate:50", batch=True)
    shell.onecmd("dir ou:pascal.area", batch=True)
    shell.onecmd("ex ou:pascal.area", batch=True)


def test_apple_single():
    info = ProDOSFileInfo(0xFF, 0x34, 0x5678)
    data = b"Hello, world!"
    resource = b"Resource"
    apple_single = encode_apple_single(info, data, resource)
    data2, resource2, info2 = decode_apple_single(apple_single)
    assert data == data2
    assert resource == resource2
    assert info.access == info2.access
    assert info.file_type == info2.file_type
    assert info.aux_type == info2.aux_type
    apple_single3 = encode_apple_single(info, data)
    data3, resource3, info3 = decode_apple_single(apple_single3)
    assert data == data3
    assert resource3 is None
    assert info == info3

    shell = Shell(verbose=True)
    shell.onecmd(f"create {DSK}.mo /allocate:2000", batch=True)
    shell.onecmd(f"init /prodos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /prodos {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')
    shell.onecmd("copy tests/dsk/ciao.apple2 ou:", batch=True)
    test1 = fs.get_file_entry("ciao.apple2")
    assert test1.storage_type == 0x2
    assert test1.prodos_file_type == 0x6
    assert test1.aux_type == 8192

    fs.write_bytes(fullname="test.extended", content=apple_single)
    test2 = fs.get_file_entry("test.extended")
    assert test2.storage_type == 0x5
    assert test2.prodos_file_type == 0x34
    assert test2.aux_type == 0x5678
    assert data == fs.read_bytes("test.extended/data.fork")
    assert resource == fs.read_bytes("test.extended/resource.fork")



import pytest

from xferx.apple2.appledosfs import AppleDOSFilesystem
from xferx.apple2.commons import (
    ProDOSFileInfo,
    decode_apple_single,
    encode_apple_single,
)
from xferx.shell import Shell

DSK = "tests/dsk/appledos.dsk"


def test_appledos():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /appledos {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, AppleDOSFilesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir/brief t:", batch=True)
    shell.onecmd("dir/brief t:*.txt", batch=True)
    shell.onecmd("dir/brief t:*.notfound", batch=True)
    shell.onecmd("type t:1.txt", batch=True)

    x = fs.read_bytes("50.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("*.TXT"))
    assert len(l) == 10
    for x in l:
        assert not x.is_empty
        assert str(x)


def test_appledos_init():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /appledos {DSK}", batch=True)
    shell.onecmd(f"create /allocate:280 {DSK}.mo", batch=True)
    shell.onecmd(f"init /appledos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /appledos {DSK}.mo", batch=True)
    shell.onecmd("dir ou:", batch=True)
    shell.onecmd("copy/type:t t:*.txt ou:", batch=True)
    fs = shell.volumes.get('OU')

    x1 = fs.read_bytes("50.txt")
    x1 = x1.rstrip(b"\0")
    assert len(x1) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x1

    with pytest.raises(Exception):
        shell.onecmd("delete ou:aaa", batch=True)
    shell.onecmd("delete ou:50.txt", batch=True)
    fs.read_bytes("10.txt")
    with pytest.raises(FileNotFoundError):
        fs.read_bytes("50.txt")

    # Test init mounted volume
    shell.onecmd("init ou:", batch=True)
    with pytest.raises(Exception):
        fs.read_bytes("10.txt")


def test_appledos_init_non_standard():
    shell = Shell(verbose=True)
    shell.onecmd(f"create {DSK}.mo /allocate:505", batch=True)
    shell.onecmd(f"init /appledos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /appledos {DSK}.mo", batch=True)


def test_apple_single():
    info = ProDOSFileInfo(0xFF, 0x34, 0x5678)
    data = b"Hello, world!"
    apple_single = encode_apple_single(info, data)
    data2, _, info2 = decode_apple_single(apple_single)
    assert info.access == info2.access
    assert info.file_type == info2.file_type
    assert info.aux_type == info2.aux_type

    shell = Shell(verbose=True)
    shell.onecmd(f"create {DSK}.mo /allocate:280", batch=True)
    shell.onecmd(f"init /appledos {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /appledos {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')

    shell.onecmd("copy tests/dsk/ciao.apple2 ou:", batch=True)
    test1 = fs.get_file_entry("ciao.apple2")
    assert test1.file_type == "B"
    apple_single3 = fs.read_bytes("ciao.apple2")
    _, _, info3 = decode_apple_single(apple_single3)
    assert info3.file_type == 0x6
    assert info3.aux_type == 0x2000



from xferx.pdp7.unix0fs import (
    V0_FIRST_INODE_BLOCK,
    V0_INODE_BLOCKS,
    V0_ROOT_INODE,
    UNIX0Filesystem,
    UNIX0FreeStorageMap,
)
from xferx.shell import Shell

DSK = "tests/dsk/unixv0.dsk"


def test_unix0_read():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /unix0 {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, UNIX0Filesystem)
    assert fs.version == 0

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir t:/", batch=True)
    shell.onecmd("dir t:/system/", batch=True)
    shell.onecmd("type t:/system/password", batch=True)

    x = fs.read_text("dd/data/9k")
    assert x.startswith("|")

    l = list(fs.entries_list)
    filenames = [x.filename for x in l if not x.is_empty]
    assert "dd" in filenames
    assert "system" in filenames

    entry = fs.get_file_entry("/test/a")
    assert not entry.inode.is_large

    entry = fs.get_file_entry("/test/b")
    assert not entry.inode.is_large

    entry = fs.get_file_entry("/test/c")
    assert entry.inode.is_large


def test_unix0_write():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /unix0 {DSK}.mo", batch=True)
    fs = shell.volumes.get('OU')

    i = fs.get_inode("/")
    assert i.inode_num == V0_ROOT_INODE
    i2 = fs.read_inode(i.inode_num)
    assert i == i2
    i2.write()
    i3 = fs.read_inode(i.inode_num)
    assert i3 == i2

    free_block_list = UNIX0FreeStorageMap.read(fs)
    for i in range(V0_FIRST_INODE_BLOCK, V0_FIRST_INODE_BLOCK + V0_INODE_BLOCKS):  # Blocks 2 to 711 contain the inodes
        assert not free_block_list.is_free(i)
        # free_block_list.set_free(i)
    # print(free_block_list.free_blocks)
    print(f"{free_block_list.used()} blocks used {free_block_list.free()} blocks free")

    free_block_list.write()

    free_block_list2 = UNIX0FreeStorageMap.read(fs)
    assert free_block_list2 == free_block_list
    assert free_block_list2.used() == free_block_list.used()
    tmp = free_block_list2.allocate(10)
    assert free_block_list2 != free_block_list
    assert free_block_list2.used() == free_block_list.used() + 10
    assert free_block_list2.free() == free_block_list.free() - 10
    free_block_list2.write()

    free_block_list3 = UNIX0FreeStorageMap.read(fs)
    for block_number in tmp:
        assert not free_block_list3.is_free(block_number)
        free_block_list3.set_free(block_number)
    free_block_list3.write()
    assert free_block_list3.used() == free_block_list.used()
    assert free_block_list3.free() == free_block_list.free()



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import sys
import typing as t
from abc import ABC, abstractmethod
from datetime import date
from enum import Enum

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, IMAGE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.block_18bit import (
    BlockDevice18Bit,
    from_18bit_words_to_bytes,
    from_bytes_to_18bit_words,
)
from .codes import (
    BAUDOT_TO_ASCII,
    fiodec_to_str,
    read_baudot_string,
    str_to_baudot,
    str_to_fiodec,
)

__all__ = [
    "DECSysFile",
    "DECSysDirectoryEntry",
    "DECSysFilesystem",
]


# DECSys-7 Operating Manual
# http://bitsavers.informatik.uni-stuttgart.de/pdf/dec/pdp7/DEC-07-SDDA-D_DECSYS7_Nov66.pdf

# Technical Notes on DECsys
# https://simh.trailing-edge.com/docs/decsys.pdf

DECTAPE_BLOCKS = 384  # Number of blocks on a DECtape
WORDS_PER_BLOCK = 256  # Number of words per block
LINKED_FILE_WORDS_PER_BLOCK = 254  # Number of words per block in a linked file

TAPE_LABEL_BLOCK = 1  # Tape label block number
PROGRAM_DIRECTORY_BLOCK = 2  # Program directory block number
LIBRARY_DIRECTORY_BLOCK = 3  # Library directory block number
KEYBOARD_MONITOR_BLOCK = 4  # Keyboard monitor block number
KEYBOARD_MONITOR_SIZE = 3  # Keyboard monitor size in blocks
FIRST_FILE_BLOCK = KEYBOARD_MONITOR_BLOCK + KEYBOARD_MONITOR_SIZE  # First block usable for files
LAST_FILE_BLOCK = DECTAPE_BLOCKS - 2  # Last block usable for files

SYSTEM_ENTRY_SIZE = 5  # System program entry size, in words
WORKING_ENTRY_SIZE = 6  # Working program entry size, in words


# File types
class FileType(Enum):
    SYSTEM = 1  # System program
    WORKING = 2  # Working program
    LIBRARY = 3  # Library program
    FORTRAN = 20  # Working program - Fortran source code
    ASSEMBLER = 21  # Working program - Assembler source code
    BINARY = 22  # Working program - Binary code

    def __str__(self) -> str:
        return self.name

    @property
    def short(self) -> str:
        return self.name[0]

    @property
    def file_mode(self) -> str:
        if self in [FileType.FORTRAN, FileType.ASSEMBLER]:
            return ASCII
        else:
            return IMAGE

    @classmethod
    def from_str(cls, value: str) -> "FileType":
        value = value.upper()
        for x in cls:
            if x.name.startswith(value):
                return x
        raise ValueError(f"Invalid file type: {value}")


def decsys_canonical_filename(fullname: str, wildcard: bool = False) -> str:
    """
    Generate the canonical DECSys filename
    """
    chars = set(BAUDOT_TO_ASCII.values())
    if wildcard:
        chars.add("*")
    return "".join([x for x in fullname.upper().strip() if x in chars])


def decsys_split_fullname(fullname: str, wildcard: bool = True) -> t.Tuple[t.Optional[FileType], str]:
    """
    Split file type and filename
    (e.g. "F,HELLO" -> FileType.FORTRAN, "HELLO")
    """
    if "," in fullname:
        file_type, fullname = fullname.split(",", 1)
        raw_file_type: t.Optional[FileType] = FileType.from_str(file_type)  # type: ignore
    else:
        raw_file_type = None
    return raw_file_type, decsys_canonical_filename(fullname, wildcard=wildcard)


def oct_dump(words: t.List[int], words_per_line: int = 8) -> None:
    """
    Display contents in octal, fiodec and baudot
    """
    for i in range(0, len(words), words_per_line):
        line = words[i : i + words_per_line]
        fiodec = "".join([x if 32 <= ord(x) <= 126 else "." for x in fiodec_to_str(line)])
        baudot = "".join([x if 32 <= ord(x) <= 126 else "." for x in read_baudot_string(line)[0]])
        oct_str = " ".join([f"{x:06o}" for x in line])
        sys.stdout.write(f"{i:08o}   {oct_str.ljust(5 * words_per_line)}  {fiodec:24}  {baudot}\n")


class DECSysFile(AbstractFile):
    entry: "DECSysDirectoryEntry"
    file_mode: str  # ASCII of IMAGE
    closed: bool

    def __init__(self, entry: "DECSysDirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        if file_mode is not None:
            self.file_mode = file_mode
        else:
            self.file_mode = entry.raw_file_type.file_mode
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if self.entry.raw_file_type == FileType.SYSTEM:
            # Contiguous files - the first block starts with:
            # word 0 - two's complement of number of words
            # word 1 - initial load address – 1
            num_words = 0
            data = bytearray()
            for i, next_block_number in enumerate(self.entry.get_blocks()):
                if i >= block_number:
                    words = self.entry.directory.fs.read_words_block(next_block_number)
                    if i == 0:  # first block
                        num_words_comp = words[0]
                        num_words = 0x40000 - num_words_comp
                        words = words[2:]  # Skip the first 2 words
                    words = words[:num_words]
                    num_words -= len(words)
                    t = from_18bit_words_to_bytes(words, self.file_mode)
                    data.extend(t)
                    number_of_blocks -= 1
                    if number_of_blocks == 0 or num_words <= 0:
                        break
        else:
            # Linked files - each block starts with:
            # word 0 - block number of next block in the file (0 if last)
            # word 1 - two's complement of number of words used in this block
            data = bytearray()
            for i, next_block_number in enumerate(self.entry.get_blocks()):
                if i >= block_number:
                    words = self.entry.directory.fs.read_words_block(next_block_number)
                    num_words_comp = words[1]
                    num_words = 0x40000 - num_words_comp
                    words = words[2 : 2 + num_words]  # Skip the first 2 words
                    t = from_18bit_words_to_bytes(words, self.file_mode)
                    data.extend(t)
                    number_of_blocks -= 1
                    if number_of_blocks == 0:
                        break
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if self.closed or block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words = from_bytes_to_18bit_words(buffer, self.file_mode)
        self.write_words_block(words, block_number, number_of_blocks)

    def write_words_block(
        self,
        words: t.List[int],
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if self.entry.raw_file_type == FileType.SYSTEM:
            # Contiguous files
            raise OSError(errno.EROFS, os.strerror(errno.EROFS))
        else:
            # Linked files
            blocks = self.entry.get_blocks()
            for i in range(block_number, block_number + number_of_blocks):
                next_block_number = blocks[i + 1] if i < len(blocks) - 1 else 0
                block_words = words[i * LINKED_FILE_WORDS_PER_BLOCK : (i + 1) * LINKED_FILE_WORDS_PER_BLOCK]
                num_words = len(block_words)
                num_words_comp = 0x40000 - num_words
                block_words = (
                    [next_block_number, num_words_comp]
                    + block_words
                    + ([0] * (LINKED_FILE_WORDS_PER_BLOCK - len(block_words)))
                )
                self.entry.directory.fs.write_words_block(blocks[i], block_words)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.entry.get_length()

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        if self.entry.raw_file_type == FileType.SYSTEM:
            return WORDS_PER_BLOCK * 3
        else:
            return LINKED_FILE_WORDS_PER_BLOCK * 3

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class DECSysDirectoryEntry(AbstractDirectoryEntry):
    """
    System program Directory Entry
    ------------------------------

    Word

        +-----------------------------------+
      0 | File type ( 1 = SYSTEM)           |
        +-----------------------------------+
      1 | Filename (BAUDOT)                 |
      2 |                                   |
        +-----------------------------------+
      3 | Block number                      |
        +-----------------------------------+
      4 | Starting address                  |
        +-----------------------------------+

    Working program Directory Entry
    -------------------------------

    Word

        +-----------------------------------+
      0 | File type ( 2 = WORKING)          |
        +-----------------------------------+
      1 | Filename (BAUDOT)                 |
      2 |                                   |
        +-----------------------------------+
      3 | FORTRAN Block number              |
        +-----------------------------------+
      4 | Assembler Block number            |
        +-----------------------------------+
      5 | Relocatable binary Block number   |
        +-----------------------------------+

    Library Directory Entry
    ----------------------

        +-----------------------------------+
      1 | Filename (BAUDOT)                 |
        /                                   /
      n |                                   |
        +-----------------------------------+
    n+1 | 777777                            |
        +-----------------------------------+
    n+2 | Block number                      |
        +-----------------------------------+
    n+3 | 777777                            |
        +-----------------------------------+

    """

    directory: "DECSysAbstractDirectory"
    filename: str  # Filename
    raw_file_type: FileType  # Raw file type
    block_number: int = 0  # Program / relocable binary / library block number
    fortran_block_number: int = 0  # Fortran block number
    assembler_block_number: int = 0  # Assembler block number
    starting_address: int = 0  # Starting address
    entry_length: int  # Entry length, in words (5 for system, 6 for working programs)

    def __init__(self, directory: "DECSysAbstractDirectory"):
        self.directory = directory

    @classmethod
    def read(cls, directory: "DECSysAbstractDirectory", words: t.List[int], position: int) -> "DECSysDirectoryEntry":
        if words[position] == FileType.SYSTEM.value:
            # System program
            return SystemDirectoryEntry.read(directory, words, position)
        else:
            # Working program
            return WorkingDirectoryEntry.read(directory, words, position)

    @abstractmethod
    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        pass

    @property
    def fullname(self) -> str:
        """Type,Filename"""
        return f"{self.raw_file_type.short},{self.filename}"

    @property
    def basename(self) -> str:
        """Filename"""
        return self.filename

    @abstractmethod
    def get_blocks(self, file_type: t.Optional[FileType] = None) -> t.List[int]:
        """
        Get the blocks used by the file
        """
        pass

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return len(self.get_blocks())

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return WORDS_PER_BLOCK * 3

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        l = len(self.directory.entries)
        self.directory.entries = [x for x in self.directory.entries if x.filename != self.filename]
        if len(self.directory.entries) < l:
            self.directory.write()
            return True
        else:
            return False

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.directory.write()
        return True

    def deallocate(self) -> None:
        """
        Deallocate the blocks used by the file
        """
        if self.raw_file_type == FileType.FORTRAN:
            self.fortran_block_number = 0
        elif self.raw_file_type == FileType.ASSEMBLER:
            self.assembler_block_number = 0
        else:  # FileType.BINARY, FileType.LIBRARY
            self.block_number = 0
        # Write the directory entry
        self.directory.write()

    def allocate(self, number_of_blocks: int) -> None:
        """
        Allocate blocks for the file
        """
        # Allocate space
        blocks, first_free_block = self.directory.fs.allocate_space(number_of_blocks)
        # Update the file entry
        if self.raw_file_type == FileType.FORTRAN:
            self.fortran_block_number = blocks[0] if blocks else 0
        elif self.raw_file_type == FileType.ASSEMBLER:
            self.assembler_block_number = blocks[0] if blocks else 0
        else:  # FileType.BINARY, FileType.LIBRARY
            self.block_number = blocks[0] if blocks else 0
        # Write the directory and update the first free block
        if isinstance(self.directory, ProgramDirectory):
            self.directory.first_free_block = first_free_block
            self.directory.write()
        else:
            self.directory.write()
            directory = ProgramDirectory(self.directory.fs)
            directory.first_free_block = first_free_block
            directory.write()

    @property
    def file_type(self) -> t.Optional[str]:
        """File type"""
        return str(self.raw_file_type)

    def open(self, file_type: t.Optional[str] = None) -> DECSysFile:
        """
        Open a file
        """
        return DECSysFile(self, file_type)

    def read_bytes(self, file_type: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        if file_type is None:
            if self.raw_file_type in [FileType.FORTRAN, FileType.ASSEMBLER]:
                file_type = ASCII
            else:
                file_type = IMAGE
        # Always read the file as IMAGE
        f = self.open(IMAGE)
        try:
            data = f.read_block(0, READ_FILE_FULL)
            if file_type == IMAGE:
                return data
            else:
                # Convert IMAGE => words => ASCII
                words = from_bytes_to_18bit_words(data, file_type=IMAGE)
                return fiodec_to_str(words).encode('ascii')
        finally:
            f.close()


class SystemDirectoryEntry(DECSysDirectoryEntry):
    """
    System program Directory Entry
    ------------------------------

    Word

        +-----------------------------------+
      0 | File type ( 1 = SYSTEM)           |
        +-----------------------------------+
      1 | Filename (BAUDOT)                 |
      2 |                                   |
        +-----------------------------------+
      3 | Block number                      |
        +-----------------------------------+
      4 | Starting address                  |
        +-----------------------------------+
    """

    @classmethod
    def read(cls, directory: "DECSysAbstractDirectory", words: t.List[int], position: int) -> "SystemDirectoryEntry":
        self = cls(directory)
        raw_filename = words[position + 1 : position + 3]
        self.filename, _ = read_baudot_string(raw_filename, 0)
        self.block_number = words[position + 3]
        self.starting_address = words[position + 4] - 1
        self.entry_length = SYSTEM_ENTRY_SIZE
        self.raw_file_type = FileType.SYSTEM
        return self

    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        return (
            [self.raw_file_type.value]
            + str_to_baudot(self.filename, length=2)
            + [self.block_number, self.starting_address + 1]
        )

    def get_blocks(self, file_type: t.Optional[FileType] = None) -> t.List[int]:
        """
        System files are stored as contiguous files
        word 0 - two's complement of number of words
        word 1 - initial load address – 1
        If the last word of the file is != 0o777777, the file continues in the next block
        """
        block_number = self.block_number
        if block_number == 0:
            return []
        num_blocks = 0
        last_word = 0
        while last_word != 0o777777:
            buffer = self.directory.fs.read_words_block(block_number + num_blocks)
            num_words = 0x40000 - buffer[0]
            t_num_blocks = math.ceil((num_words + 2) / 256)
            if t_num_blocks > 1:
                # Read the last block to check the last word
                buffer = self.directory.fs.read_words_block(block_number + num_blocks + t_num_blocks - 1)
            last_word = buffer[(num_words + 2) % 256]
            num_blocks += t_num_blocks
        return list(range(block_number, block_number + num_blocks))

    def __str__(self) -> str:
        return f"{self.filename:<15} Type: {self.file_type}  Block: {self.block_number:04}  Addr: {self.starting_address:>06}"


class WorkingDirectoryEntry(DECSysDirectoryEntry):
    """
    Working program Directory Entry
    -------------------------------

    Word

        +-----------------------------------+
      0 | File type ( 2 = WORKING)          |
        +-----------------------------------+
      1 | Filename (BAUDOT)                 |
      2 |                                   |
        +-----------------------------------+
      3 | FORTRAN Block number              |
        +-----------------------------------+
      4 | Assembler Block number            |
        +-----------------------------------+
      5 | Relocatable binary Block number   |
        +-----------------------------------+
    """

    @classmethod
    def read(cls, directory: "DECSysAbstractDirectory", words: t.List[int], position: int) -> "WorkingDirectoryEntry":
        self = cls(directory)
        raw_filename = words[position + 1 : position + 3]
        self.filename, _ = read_baudot_string(raw_filename, 0)
        self.fortran_block_number = words[position + 3]
        self.assembler_block_number = words[position + 4]
        self.block_number = words[position + 5]
        self.entry_length = WORKING_ENTRY_SIZE
        self.raw_file_type = FileType.WORKING
        return self

    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        return (
            [FileType.WORKING.value]
            + str_to_baudot(self.filename, length=2)
            + [self.fortran_block_number, self.assembler_block_number, self.block_number]
        )

    def get_blocks(self, file_type: t.Optional[FileType] = None) -> t.List[int]:
        """
        Text files and library files are stored as linked files
        word 0 - block number of next block in the file (0 if last)
        word 1 - two's complement of number of words used in this block
        """
        # For working files, get the FORTRAN, Assembler of Binary fork
        file_type = file_type or self.raw_file_type
        if file_type == FileType.FORTRAN:
            block_number = self.fortran_block_number
        elif file_type == FileType.ASSEMBLER:
            block_number = self.assembler_block_number
        else:
            block_number = self.block_number
        next_block_number = block_number
        blocks = []
        while next_block_number:
            blocks.append(next_block_number)
            next_block_number = self.directory.fs.read_words_block(next_block_number)[0]
        return blocks

    def __str__(self) -> str:
        if self.raw_file_type == FileType.FORTRAN:  # Fortran source code
            addr = f"Block: {self.fortran_block_number:04}"
        elif self.raw_file_type == FileType.ASSEMBLER:  # Assembler source code
            addr = f"Block: {self.assembler_block_number:04}"
        elif self.raw_file_type == FileType.BINARY:  # Binary code
            addr = f"Block: {self.block_number:04}"
        elif self.raw_file_type == FileType.WORKING:  # Working program
            addr = f"Fortran: {self.fortran_block_number:04}  Asm: {self.assembler_block_number:04}  Binary: {self.block_number:04}"
        else:
            addr = ""
        return f"{self.filename:<15} Type: {self.file_type}  {addr}"


class LibraryDirectoryEntry(DECSysDirectoryEntry):
    """
    Library Directory Entry
    ----------------------

        +-----------------------------------+
      1 | Filename (BAUDOT)                 |
        /                                   /
      n |                                   |
        +-----------------------------------+
    n+1 | 777777                            |
        +-----------------------------------+
    n+2 | Block number                      |
        +-----------------------------------+
    n+3 | 777777                            |
        +-----------------------------------+
    """

    @classmethod
    def read(cls, directory: "DECSysAbstractDirectory", words: t.List[int], position: int) -> "LibraryDirectoryEntry":
        assert isinstance(directory, LibraryDirectory)
        # Directory entries names are variable length,
        # depending on the number of entry points
        self = cls(directory)
        self.filename, pp = read_baudot_string(words, position)
        self.block_number = words[pp + 1]
        self.entry_length = pp - position + 3
        self.raw_file_type = FileType.LIBRARY
        return self

    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        # Directory entries names are variable length,
        # depending on the number of entry points
        words = str_to_baudot(self.filename) + [
            0o777777,
            self.block_number,
            0o777777,
        ]
        self.entry_length = len(words)
        return words

    def get_blocks(self, file_type: t.Optional[FileType] = None) -> t.List[int]:
        """
        Text files and library files are stored as linked files
        word 0 - block number of next block in the file (0 if last)
        word 1 - two's complement of number of words used in this block
        """
        next_block_number = self.block_number
        blocks = []
        while next_block_number:
            blocks.append(next_block_number)
            next_block_number = self.directory.fs.read_words_block(next_block_number)[0]
        return blocks

    def __str__(self) -> str:
        return f"{self.filename:<15} Type: {self.file_type}  Block: {self.block_number:04}"


class DECSysAbstractDirectory(ABC):
    """
    Abstract Directory - the two concrete directories
    are Program Directory and Library Directory
    """

    fs: "DECSysFilesystem"
    entries: t.List["DECSysDirectoryEntry"]

    def __init__(self, fs: "DECSysFilesystem"):
        self.fs = fs

    @classmethod
    @abstractmethod
    def read(cls, fs: "DECSysFilesystem") -> "DECSysAbstractDirectory":
        """
        Read the Directory from disk
        """
        pass

    @abstractmethod
    def write(self) -> None:
        """
        Write the Directory to the disk
        """
        pass


class ProgramDirectory(DECSysAbstractDirectory):
    """
    Program Directory (Block 2)

    Word

        +-----------------------------------+
      0 | Directory Length (words)          |
        +-----------------------------------+
      1 | Program Directory entries         |
        /                                   /
        |                                   |
        +-----------------------------------+
    255 | First free block number           |
        +-----------------------------------+
    """

    first_free_block: int  # First free block number

    @classmethod
    def read(cls, fs: "DECSysFilesystem") -> "ProgramDirectory":
        """
        Read Program Directory entries

        https://simh.trailing-edge.com/docs/decsys.pdf  Pag 2
        http://bitsavers.informatik.uni-stuttgart.de/pdf/dec/pdp7/DEC-07-SDDA-D_DECSYS7_Nov66.pdf  Pag 12
        """
        self = ProgramDirectory(fs)
        self.entries = []
        words = self.fs.read_words_block(PROGRAM_DIRECTORY_BLOCK)
        dir_length = words[0]  # Directory length, in words
        self.first_free_block = words[255]  # First free block number
        position = 1
        while position < len(words) - 5 and position < dir_length:
            entry = DECSysDirectoryEntry.read(self, words, position)
            position += entry.entry_length
            self.entries.append(entry)
        return self

    def write(self) -> None:
        """
        Write the Program Directory to the disk
        """
        words = [0]
        for entry in self.entries:
            words += entry.to_words()
        words[0] = len(words) - 1  # Directory length, in words
        words += [0] * (255 - len(words))  # pad
        words += [self.first_free_block]  # First free block number
        self.fs.write_words_block(PROGRAM_DIRECTORY_BLOCK, words)


class LibraryDirectory(DECSysAbstractDirectory):
    """
    Library Directory (Block 3)

    Word

        +-----------------------------------+
      0 | Directory Length (words)          |
        +-----------------------------------+
      1 | Library Directory entries         |
        /                                   /
        |                                   |
        +-----------------------------------+
    """

    @classmethod
    def read(cls, fs: "DECSysFilesystem") -> "LibraryDirectory":
        """
        Read Library Directory entries

        https://simh.trailing-edge.com/docs/decsys.pdf  Pag 2
        http://bitsavers.informatik.uni-stuttgart.de/pdf/dec/pdp7/DEC-07-SDDA-D_DECSYS7_Nov66.pdf  Pag 12
        """
        self = LibraryDirectory(fs)
        self.entries = []
        words = self.fs.read_words_block(LIBRARY_DIRECTORY_BLOCK)
        dir_length = words[0]  # Directory length, in words
        position = 1
        while position < len(words) - 5 and position < dir_length:
            entry = LibraryDirectoryEntry.read(self, words, position)
            position += entry.entry_length
            self.entries.append(entry)
        return self

    def write(self) -> None:
        """
        Write the Library Directory to the disk
        """
        words = [0]
        for entry in self.entries:
            words += entry.to_words()
        words[0] = len(words) - 1  # Directory length, in words
        words += [0] * (256 - len(words))  # pad
        self.fs.write_words_block(LIBRARY_DIRECTORY_BLOCK, words)


class DECSysFilesystem(AbstractFilesystem):
    """

    DECtapes prepared for the DECSYS-7 user contains
    384 blocks of 256 18-bit words each.

    Dectape
    =======

    Block

            +-------------------+
      0     | Unused            |
            +-------------------+
      1     | Tape label        |
            +-------------------+
      2     | Program directory |
            +-------------------+
      3     | Library directory |
            +-------------------+
      4     | Keyboard          |
      6     | Monitor           |
            +-------------------+
      7     | Files             |
            /                   /
    382     |                   |
            +-------------------+
    383     | Unused            |
            +-------------------+

    """

    fs_name = "decsys"
    fs_description = "PDP-7 DECSys"
    fs_platforms = ["pdp-7"]
    dev: BlockDevice18Bit

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = BlockDevice18Bit(file_or_device, words_per_block=WORDS_PER_BLOCK)
        elif isinstance(file_or_device, BlockDevice18Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "DECSysFilesystem":
        """
        Mount the filesystem from a file or device
        """
        self = cls(file_or_dev)
        if strict:
            if self.get_size() // WORDS_PER_BLOCK // 4 != DECTAPE_BLOCKS:
                raise OSError(errno.EINVAL, "Invalid DECSys tape size")
        return self

    def read_words_block(
        self,
        block_number: int,
    ) -> t.List[int]:
        """
        Read a 256 bytes block as 18bit words
        """
        return self.dev.read_words_block(block_number)

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 18bit words as a block
        """
        self.dev.write_words_block(block_number, words)

    def read_tape_label(self) -> t.Tuple[str, str]:
        """
        Read tape label

        The tape label contains alphanumeric strings which identify the tape name and
        the date on which it was updated.

        The labels are of two Baudot strings, padded with 0’s to an 18b boundary and
        terminated by a word of all ones (0o777777).

        https://simh.trailing-edge.com/docs/decsys.pdf  Pag 2
        http://bitsavers.informatik.uni-stuttgart.de/pdf/dec/pdp7/DEC-07-SDDA-D_DECSYS7_Nov66.pdf  Pag 12
        """
        words = self.read_words_block(TAPE_LABEL_BLOCK)
        label1, position = read_baudot_string(words, 0)
        label2, _ = read_baudot_string(words, position + 1)
        return label1, label2

    def read_dir_entries(self) -> t.Iterator["DECSysDirectoryEntry"]:
        """
        Read directory entries

        https://simh.trailing-edge.com/docs/decsys.pdf  Pag 2
        http://bitsavers.informatik.uni-stuttgart.de/pdf/dec/pdp7/DEC-07-SDDA-D_DECSYS7_Nov66.pdf  Pag 12
        """
        program_directory = ProgramDirectory.read(self)
        yield from program_directory.entries
        library_directory = LibraryDirectory.read(self)
        yield from library_directory.entries

    @property
    def entries_list(self) -> t.Iterator["DECSysDirectoryEntry"]:
        yield from self.read_dir_entries()

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["DECSysDirectoryEntry"]:
        if pattern:
            raw_file_type, pattern = decsys_split_fullname(pattern, wildcard=True)
        else:
            raw_file_type = None
        for entry in self.read_dir_entries():
            if filename_match(entry.basename, pattern, wildcard):
                # Filter by file type
                if raw_file_type is not None:
                    if raw_file_type in [FileType.ASSEMBLER, FileType.FORTRAN, FileType.BINARY]:
                        if entry.raw_file_type != FileType.WORKING:
                            continue
                    elif raw_file_type != entry.raw_file_type:
                        continue
                    # Assign the file type to the entry
                    entry.raw_file_type = raw_file_type
                yield entry

    def get_file_entry(self, fullname: str) -> DECSysDirectoryEntry:
        """
        Get the directory entry for a file
        """
        # Split file type and filename (e.g. "F,HELLO" -> FileType.FORTRAN, "HELLO")
        raw_file_type, filename = decsys_split_fullname(fullname)
        for entry in self.read_dir_entries():
            if entry.filename == filename:
                # Filter by file type
                if raw_file_type is not None:
                    if raw_file_type in [FileType.ASSEMBLER, FileType.FORTRAN, FileType.BINARY]:
                        if entry.raw_file_type != FileType.WORKING:
                            continue
                    elif raw_file_type != entry.raw_file_type:
                        continue
                    # Assign the file type to the entry
                    entry.raw_file_type = raw_file_type
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def get_allocated_blocks(self) -> t.List[int]:
        """
        Get the list of allocated blocks
        """
        blocks = list(range(0, FIRST_FILE_BLOCK))
        for entry in self.read_dir_entries():
            if entry.raw_file_type == FileType.WORKING:
                for raw_file_type in [FileType.FORTRAN, FileType.ASSEMBLER, FileType.BINARY]:
                    blocks.extend(entry.get_blocks(raw_file_type))
            else:
                blocks.extend(entry.get_blocks())
        return blocks

    def allocate_space(self, number_of_blocks: int) -> t.Tuple[t.List[int], int]:
        """
        Allocate blocks for a file
        Return the list of allocated blocks and the first free block number
        """
        allocated_blocks = self.get_allocated_blocks()
        blocks = []
        for block in range(FIRST_FILE_BLOCK, LAST_FILE_BLOCK + 1):
            if block not in allocated_blocks:
                blocks.append(block)
                if len(blocks) == number_of_blocks:
                    break
        if len(blocks) < number_of_blocks:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        # Write the linked blocks
        for i in range(0, len(blocks)):
            next_block_number = blocks[i + 1] if i < len(blocks) - 1 else 0
            block_words = [8] * LINKED_FILE_WORDS_PER_BLOCK
            num_words = len(block_words)
            num_words_comp = 0x40000 - num_words
            block_words = (
                [next_block_number, num_words_comp]
                + block_words
                + ([0] * (LINKED_FILE_WORDS_PER_BLOCK - len(block_words)))
            )
            self.write_words_block(blocks[i], block_words)
        first_free_block = max(allocated_blocks + blocks) + 1
        return blocks, first_free_block

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[DECSysDirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        raw_file_type, filename = decsys_split_fullname(fullname)
        if raw_file_type is None:
            raise ValueError("Specify the file type as FILE_TYPE,FILENAME")
        if raw_file_type == FileType.SYSTEM:
            raise ValueError("SYSTEM files are not supported")
        if raw_file_type == FileType.LIBRARY:
            raw_file_type = FileType.BINARY
        # If the file already exists, deallocate the blocks
        try:
            entry: DECSysDirectoryEntry = self.get_file_entry(fullname)  # type: ignore
            entry.deallocate()
        except FileNotFoundError:
            # Create a new entry
            directory = LibraryDirectory(self) if raw_file_type == FileType.LIBRARY else ProgramDirectory.read(self)
            if raw_file_type == FileType.SYSTEM:
                # entry = SystemDirectoryEntry(directory)
                raise ValueError("SYSTEM files are not supported")
            elif raw_file_type == FileType.LIBRARY:
                entry = LibraryDirectoryEntry(directory)
            else:
                entry = WorkingDirectoryEntry(directory)
            entry.filename = filename
            entry.raw_file_type = raw_file_type
            directory.entries.append(entry)
        # Allocate space
        entry.allocate(number_of_blocks)
        return entry

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        words = str_to_fiodec(content.decode('ascii'))
        number_of_blocks = int(math.ceil(len(content) / LINKED_FILE_WORDS_PER_BLOCK))
        entry = self.create_file(fullname, number_of_blocks, creation_date, file_type)
        if entry is not None:
            f = entry.open(file_type)
            try:
                f.write_words_block(words, block_number=0, number_of_blocks=number_of_blocks)
            finally:
                f.close()

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        entries = self.filter_entries_list(pattern, wildcard=True)
        if not entries:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), pattern)
        # Read tape label
        if not options.get("brief"):
            label1, label2 = self.read_tape_label()
            sys.stdout.write(f"{label1}  {label2}\n")
        for x in entries:
            if options.get("brief"):
                sys.stdout.write(f"{x.basename}\n")
            elif x.raw_file_type == FileType.SYSTEM:
                sys.stdout.write(f"{x.basename} S {x.block_number:04}\n")
            elif x.raw_file_type == FileType.LIBRARY:
                sys.stdout.write(f"{x.basename}, L {x.block_number:04}\n")
            else:
                # For working programs, starting tape-block numbers are shown for,
                # the FORTRAN source, the assembly source, the relocatable binary
                sys.stdout.write(
                    f"{x.basename} W {x.fortran_block_number:04},{x.assembler_block_number:04},{x.block_number:04}\n"
                )

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("bitmap"):
            allocated_blocks = sorted(self.get_allocated_blocks())
            tmp = ", ".join([f"{x:04}" for x in allocated_blocks])
            sys.stdout.write(f"Allocated Blocks: {tmp}\n")
        elif arg:
            entries = self.filter_entries_list(arg, wildcard=True)
            for entry in entries:
                sys.stdout.write(f"Filename:                 {entry.filename}\n")
                sys.stdout.write(f"File type:                {entry.file_type}\n")
                if entry.raw_file_type == FileType.SYSTEM:
                    sys.stdout.write(f"Blocks:                   {entry.get_blocks()}\n")
                    sys.stdout.write(f"Starting address:         {entry.starting_address:>06}\n")
                elif entry.raw_file_type == FileType.LIBRARY:
                    sys.stdout.write(f"Blocks:                   {entry.get_blocks()}\n")
                else:
                    if entry.raw_file_type in (FileType.WORKING, FileType.FORTRAN):
                        sys.stdout.write(f"FORTRAN Blocks:           {entry.get_blocks(FileType.FORTRAN)}\n")
                    if entry.raw_file_type in (FileType.WORKING, FileType.ASSEMBLER):
                        sys.stdout.write(f"Assembler Blocks:         {entry.get_blocks(FileType.ASSEMBLER)}\n")
                    if entry.raw_file_type in (FileType.WORKING, FileType.BINARY):
                        sys.stdout.write(f"Binary Blocks:            {entry.get_blocks(FileType.BINARY)}\n")
                sys.stdout.write("\n")
        else:
            label1, label2 = self.read_tape_label()
            sys.stdout.write(f"Label 1:                  {label1}\n")
            sys.stdout.write(f"Label 2:                  {label2}\n")
            program_directory = ProgramDirectory.read(self)
            sys.stdout.write(f"First free block number:  {program_directory.first_free_block:04}\n\n")
            sys.stdout.write("Filename        Type     Address  Length\n")
            sys.stdout.write("--------        ----     -------  ------\n")
            for entry in self.read_dir_entries():
                if entry.raw_file_type == FileType.WORKING:
                    for raw_file_type in [FileType.FORTRAN, FileType.ASSEMBLER, FileType.BINARY]:
                        blocks = entry.get_blocks(raw_file_type)
                        length = len(blocks)
                        file_type = str(raw_file_type)
                        block = blocks[0] if blocks else 0
                        filename = f"{file_type[0]},{entry.filename}"
                        sys.stdout.write(f"{filename:<15} {file_type:<10}  {block:04}  {length}\n")
                else:
                    file_type = str(entry.raw_file_type)
                    blocks = entry.get_blocks()
                    length = len(blocks)
                    block = blocks[0] if blocks else 0
                    filename = f"{file_type[0]},{entry.filename}"
                    sys.stdout.write(f"{filename:<15} {file_type:<10}  {block:04}  {length}\n")

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        if fullname:
            entry = self.get_file_entry(fullname)
            if start is None:
                start = 0
            blocks = entry.get_blocks()
            if end is None or end > len(blocks) - 1:
                end = entry.get_length() - 1
            for block_number in range(start, end + 1):
                words = self.read_words_block(blocks[block_number])
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)
        else:
            if start is None:
                start = 0
                if end is None:  # full disk
                    end = (self.get_size() // WORDS_PER_BLOCK // 4) - 1
            elif end is None:  # one single block
                end = start
            for block_number in range(start, end + 1):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def get_pwd(self) -> str:
        return ""

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return [
            IMAGE,
            ASCII,
        ]



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import shlex
import sys
import traceback
import typing as t

from ..commons import splitdrive
from ..volumes import Volumes
from .cmds import cmds
from .commons import CommandError, add_slash, split_command_line

try:
    import readline
except:
    readline = None  # type: ignore

__all__ = [
    "Shell",
]


HISTORY_FILENAME = "~/.rt_history"
HISTORY_LENGTH = 1000


class Shell:
    verbose: bool = False  # Verbose mode
    volumes: Volumes  # Volumes manager
    prompt: str  # Command prompt
    history_file: str  # Path to the history file
    completion_matches: t.List[str]  # Matches for shell completion

    def __init__(self, verbose: bool = False):
        self.verbose = verbose
        self.volumes = Volumes()
        self.update_prompt()
        self.history_file = os.path.expanduser(HISTORY_FILENAME)
        self.completion_matches = []
        # Init readline and history
        if readline is not None:
            if sys.platform == "darwin":
                readline.parse_and_bind("bind ^I rl_complete")
            else:
                readline.parse_and_bind("tab: complete")
                readline.parse_and_bind("set bell-style none")
            readline.set_completer(self.complete)
            try:
                if self.history_file:
                    readline.set_history_length(HISTORY_LENGTH)
                    readline.read_history_file(self.history_file)
            except IOError:
                pass

    def complete(self, text: str, state: int) -> t.Optional[str]:
        """
        Return the next possible completion
        """
        if state == 0:
            rline = readline.get_line_buffer()
            line = rline.lstrip()
            begidx = readline.get_begidx() - len(rline) + len(line)
            if begidx <= 0:
                # If the line is empty, complete commands and volumes
                self.completion_matches = self.complete_commands_volumes(text)
            else:
                cmd, args = split_command_line(line)
                if cmd == "SHOW":
                    # Complete subcommands
                    self.completion_matches = self.complete_subcommands(cmd, args or "")
                else:
                    # Complete filenames
                    self.completion_matches = self.complete_filenames(text)
        try:
            return self.completion_matches[state]
        except IndexError:
            return None

    def complete_commands_volumes(self, text: str) -> t.List[str]:
        """
        Complete commands or volumes
        """
        if text[:1].islower():
            # If the first character is lowercase, complete commands in lowercase
            commands = [x.lower() for x in cmds.get_commands()]
        else:
            commands = cmds.get_commands()
        text = text.upper()
        return [x for x in commands if x.upper().startswith(text)] + [
            f"{x}:" for x in self.volumes.volumes if x.upper().startswith(text)
        ]

    def complete_subcommands(self, cmd: str, text: str) -> t.List[str]:
        """
        Complete subcommands
        """
        if text[:1].islower():
            # If the first character is lowercase, complete commands in lowercase
            commands = [x.lower() for x in cmds.get_commands()]
        else:
            commands = cmds.get_commands()
        text = f"{cmd} {text}".upper()
        return [x.split(" ", 1)[1] for x in commands if x.upper().startswith(text)]

    def complete_filenames(self, text: str) -> t.List[str]:
        """
        Complete filenames based on the current drive and directory
        """
        try:
            has_volume_id = ":" in text
            volume_id, path = splitdrive(text)
            pattern = path + "*"
            fs = self.volumes.get(volume_id)
            result: t.List[str] = []
            for x in fs.filter_entries_list(pattern):
                tmp = add_slash(fs, x.basename)
                result.append(f"{volume_id}:{tmp}" if has_volume_id else tmp)
            return result
        except Exception:
            pass  # no problem :-)
        return []

    def setup_completer(self) -> None:
        """
        Set up the readline completer
        """
        if readline is not None:
            try:
                self.old_completer = readline.get_completer()
                readline.set_completer(self.complete)
                readline.parse_and_bind("tab: complete")
            except Exception:
                pass

    def teardown_completer(self) -> None:
        """
        Restore the old completer
        """
        if readline is not None:
            try:
                readline.set_completer(self.old_completer)
            except Exception:
                pass

    def write_history(self) -> None:
        """
        Write the command history to the history file
        """
        if readline is not None and self.history_file:
            try:
                readline.set_history_length(HISTORY_LENGTH)
                readline.write_history_file(self.history_file)
            except Exception:
                pass

    def update_prompt(self) -> None:
        """
        Update the command prompt
        """
        self.prompt = "[%s] " % self.volumes.get_pwd()

    def cmd_loop(self) -> None:
        """
        Main command loop
        """
        # Set up readline completion
        self.setup_completer()
        try:
            while True:
                self.update_prompt()
                try:
                    line = input(self.prompt)
                    self.onecmd(line)
                except SystemExit:
                    break
                except EOFError:
                    break
                except KeyboardInterrupt:
                    sys.stdout.write("\n")
            # Write history file
            self.write_history()
        finally:
            # Restore the old completer
            self.teardown_completer()

    def onecmd(self, line: str, catch_exceptions: bool = True, batch: bool = False) -> None:
        """
        Execute a single command line
        """
        try:
            cmd, arg = split_command_line(line)
            if not cmd:
                # Empty line
                pass
            elif arg == ":":
                # Set default volume
                self.volumes.set_default_volume(cmd)
            else:
                # Execute the command
                func = cmds.get(cmd)
                if func is None:
                    raise CommandError("?KMON-F-Illegal command")
                args = shlex.split(arg) if arg else []
                func(self, args)
        except KeyboardInterrupt:
            sys.stdout.write("\n")
            sys.stdout.write("\n")
        except SystemExit as ex:
            # Allow SystemExit to propagate
            raise ex
        except Exception as ex:
            if not catch_exceptions:
                raise ex
            message = str(sys.exc_info()[1])
            sys.stdout.write(f"{message}\n")
            if self.verbose and not isinstance(ex, CommandError):
                traceback.print_exc()
            if batch:
                raise ex



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import typing as t

from ..commons import BLOCK_SIZE
from .abstract import AbstractDevice

if t.TYPE_CHECKING:
    from ..abstract import AbstractFile

__all__ = [
    "BlockDevice",
]


class BlockDevice(AbstractDevice):
    """
    Block device
    """

    f: "AbstractFile"
    size: int  # Block device size, in bytes

    def __init__(self, file: "AbstractFile"):
        self.f = file
        self.size = self.f.get_size()
        self.sector_size = BLOCK_SIZE

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        return self.f.read_block(block_number, number_of_blocks)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        if block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        self.f.write_block(buffer, block_number, number_of_blocks)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.f.get_size()

    def close(self) -> None:
        self.f.close()



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import io
import struct
import typing as t

from ..abstract import AbstractFile
from ..commons import BLOCK_SIZE, swap_words
from ..device.abstract import AbstractDevice
from .commons import UNIXFilesystem, UNIXInode, format_mode, iterate_long, l3tol
from .unix4fs import (
    V4_RGRP,
    V4_ROTH,
    V4_ROWN,
    V4_SGID,
    V4_STXT,
    V4_SUID,
    V4_WGRP,
    V4_WOTH,
    V4_WOWN,
    V4_XGRP,
    V4_XOTH,
    V4_XOWN,
)

__all__ = ["UNIX7Filesystem", "UNIXInode7"]

# Version 7

V7_MPB = 0o0070000  # multiplexed block special
V7_REG = 0o0100000  # regular
V7_BLK = 0o0060000  # block special
V7_DIR = 0o0040000  # directory
V7_MPC = 0o0030000  # multiplexed char special
V7_CHR = 0o0020000  # character special

V7_PERMS = [
    [(V7_BLK, "b"), (V7_DIR, "d"), (V7_CHR, "c"), (0, "-")],
    [(V4_ROWN, "r"), (0, "-")],
    [(V4_WOWN, "w"), (0, "-")],
    [(V4_SUID, "s"), (V4_XOWN, "x"), (0, "-")],
    [(V4_RGRP, "r"), (0, "-")],
    [(V4_WGRP, "w"), (0, "-")],
    [(V4_SGID, "s"), (V4_XGRP, "x"), (0, "-")],
    [(V4_ROTH, "r"), (0, "-")],
    [(V4_WOTH, "w"), (0, "-")],
    [(V4_XOTH, "x"), (0, "-")],
    [(V4_STXT, "t"), (0, " ")],
]

V7_INODE_FORMAT = "<HHHH HH 40s III"
V7_FILENAME_LEN = 14
V7_INODE_SIZE = 64
V7_NADDR = 13
V7_DIR_FORMAT = f"H{V7_FILENAME_LEN}s"
V7_ROOT_INODE = 2

V7_NICINOD = 100  # number of superblock inodes
V7_NICFREE = 50  # number of superblock free blocks
V7_SUPER_BLOCK = 1  # Superblock
V7_SUPER_BLOCK_FORMAT = f"<Hlh {V7_NICFREE}l h {V7_NICINOD}H BBBB L"

assert struct.calcsize(V7_INODE_FORMAT) == V7_INODE_SIZE


class UNIXInode7(UNIXInode):

    @classmethod
    def read(cls, fs: "UNIXFilesystem", inode_num: int, buffer: bytes, position: int = 0) -> "UNIXInode":
        self = UNIXInode7(fs)
        self.inode_num = inode_num
        (
            self.flags,  #    1 word  flags
            self.nlinks,  #   1 word  number of links to file
            self.uid,  #      1 word  user ID of owner
            self.gid,  #      1 word  group ID of owner
            sz0,  #           1 word  high word of size
            sz1,  #           1 word  low word of size
            addr,  #          40 chars disk block addresses
            self.atime,  #    1 long  time of last access
            self.mtime,  #    1 long  time of last modification
            self.ctime,  #    1 long  time created
        ) = struct.unpack_from(V7_INODE_FORMAT, buffer, position)
        self.addr = l3tol(addr, V7_NADDR)
        self.size = (sz0 << 16) + sz1
        self.atime = swap_words(self.atime)
        self.mtime = swap_words(self.mtime)
        self.ctime = swap_words(self.ctime)
        return self

    @property
    def isdir(self) -> bool:
        return (self.flags & V7_DIR) == V7_DIR

    @property
    def is_regular_file(self) -> bool:
        return not self.isdir

    @property
    def is_allocated(self) -> bool:
        return self.flags != 0

    def blocks(self) -> t.Iterator[int]:
        rem = self.get_size()
        for block_number in self.addr[:-3]:
            if block_number == 0:
                break
            rem -= self.get_block_size()
            yield block_number
        if rem > 0:
            # Indirect block
            block_number = self.addr[-3]
            if block_number != 0:
                indirect_block = self.fs.read_block(block_number)
                for n in iterate_long(indirect_block):
                    if n != 0:
                        rem -= self.get_block_size()
                        yield n
        if rem > 0:
            # Double indirect block
            block_number = self.addr[-2]
            if block_number != 0:
                double_indirect_block = self.fs.read_block(block_number)
                for d in iterate_long(double_indirect_block):
                    if d != 0:
                        indirect_block = self.fs.read_block(d)
                        for n in iterate_long(indirect_block):
                            if n != 0:
                                rem -= self.get_block_size()
                                yield n
        if rem > 0:
            # Triple indirect block
            block_number = self.addr[-1]
            if block_number != 0:
                triple_indirect_block = self.fs.read_block(block_number)
                for tmp in iterate_long(triple_indirect_block):
                    if tmp != 0:
                        double_indirect_block = self.fs.read_block(tmp)
                        for d in iterate_long(double_indirect_block):
                            if d != 0:
                                indirect_block = self.fs.read_block(d)
                                for n in iterate_long(indirect_block):
                                    if n != 0:
                                        rem -= self.get_block_size()
                                        yield n

    def examine(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Inode\n")
        buf.write(f"Inode number:          {self.inode_num:>6}\n")
        buf.write(f"Flags:                 {self.flags:>06o}\n")
        if self.isdir:
            buf.write("Type:               directory\n")
        # elif self.is_special_file:
        #     buf.write("Type:            special file\n")
        # elif self.is_large:
        #     buf.write("Type:              large file\n")
        else:
            buf.write("Type:                    file\n")
        buf.write(f"Owner user id:         {self.uid:>6}\n")
        buf.write(f"Group user id:         {self.gid:>6}\n")
        buf.write(f"Link count:            {self.nlinks:>6}\n")
        buf.write(f"Size:                  {self.size:>6}\n")
        # if self.is_large:
        #     buf.write(f"Indirect blocks:       {self.addr}\n")
        buf.write(f"Blocks:                {list(self.blocks())}\n")
        return buf.getvalue()

    def __str__(self) -> str:
        if not self.is_allocated:
            return f"{self.inode_num:>4}# ---"
        else:
            mode = format_mode(self.flags, version=self.fs.version, perms=self.fs.perms)
            return f"{self.inode_num:>4}# {self.uid:>3},{self.gid:<3} nlinks: {self.nlinks:>3} size: {self.size:>8}  {mode} flags: {self.flags:06o}"


class UNIX7Filesystem(UNIXFilesystem):
    """
    UNIX version 7 Filesystem
    """

    fs_name = "unix7"
    fs_description = "UNIX version 7"
    fs_platforms = ["pdp-11", "interdata8", "vax"]
    version: int = 7  # UNIX version
    inode_size = V7_INODE_SIZE
    dir_format = V7_DIR_FORMAT
    root_inode = V7_ROOT_INODE
    perms = V7_PERMS
    unix_inode_class = UNIXInode7

    @classmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "UNIX7Filesystem":
        self = cls(file_or_dev)
        self.pwd = "/"
        self.read_superblock()
        return self

    def read_superblock(self) -> None:
        superblock_data = self.read_block(V7_SUPER_BLOCK)
        superblock = struct.unpack_from(V7_SUPER_BLOCK_FORMAT, superblock_data, 0)
        self.inode_list_blocks = superblock[0]  # number of blocks devoted to the i-list
        self.volume_size = swap_words(superblock[1])  # size in blocks of entire volume
        self.free_blocks_in_list = superblock[2]  # number of free blocks in the free list
        self.free_blocks_list = list([swap_words(x) for x in superblock[3 : 3 + V7_NICINOD]])  # free block list
        self.free_inodes_in_list = superblock[3 + V7_NICINOD]  # number of free inodes in the inode list
        self.free_inodes_list = list(superblock[4 + V7_NICINOD : 4 + V7_NICINOD + V7_NICFREE])  # free inode list
        # _ = superblock[4 + V7_NICINOD + V7_NICFREE]  # lock during free list manipulation
        # _ = superblock[5 + V7_NICINOD + V7_NICFREE]  # lock during i-list manipulation
        # _ = superblock[6 + V7_NICINOD + V7_NICFREE]  # flag to indicate that the super-block has changed and should be written
        # _ = superblock[7 + V7_NICINOD + V7_NICFREE]  # mounted read-only flag
        # _ = swap_words(superblock[8 + V7_NICINOD + V7_NICFREE]) # last super block update
        self.inodes = (self.inode_list_blocks - 1) * (BLOCK_SIZE // self.inode_size)



from xferx.native import NativeFilesystem
from xferx.shell import Shell


def test_native():
    shell = Shell(verbose=True)
    fs = shell.volumes.get("DK")
    assert isinstance(fs, NativeFilesystem)

    shell.onecmd("dir dk:", batch=True)
    shell.onecmd("type LICENSE", batch=True)

    with open("t.tmp", "w"):
        pass

    e = fs.read_bytes("t.tmp")
    assert len(e) == 0

    f = fs.open_file("t.tmp")

    x = fs.read_bytes("LICENSE")
    assert x
    for line in x.split(b"\n"):
        f.write(line)
        f.write(b"\n")
    f.truncate(len(x))
    f.close()

    x2 = fs.read_bytes("t.tmp")
    assert x == x2

    shell.onecmd("delete t.tmp", batch=True)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import typing as t

from ..commons import BLOCK_SIZE, READ_FILE_FULL
from .block import BlockDevice

if t.TYPE_CHECKING:
    from ..abstract import AbstractFile

__all__ = [
    "RXBlockDevice",
    "RX_SECTOR_TRACK",
    "RX_TRACK_DISK",
    "RX01_SECTOR_SIZE",
    "RX02_SECTOR_SIZE",
    "RX01_SIZE",
    "RX02_SIZE",
    "get_sector_size",
    "rxfactr",
    "rxfactr_12bit",
    "rx_extract_12bit_words",
    "rx_pack_12bit_words",
]

RX_SECTOR_TRACK = 26  # sectors/track
RX_TRACK_DISK = 77  # track/disk
RX01_SECTOR_SIZE = 128  # RX01 bytes/sector
RX02_SECTOR_SIZE = 256  # RX02 bytes/sector
RX01_SIZE = RX_TRACK_DISK * RX_SECTOR_TRACK * RX01_SECTOR_SIZE  # RX01 Capacity
RX02_SIZE = RX_TRACK_DISK * RX_SECTOR_TRACK * RX02_SECTOR_SIZE  # RX02 Capacity

# Interleave tables for RX01 in 12-bit mode
RX01_INTERLEAVE_12B = [
    (t * RX_SECTOR_TRACK) + s - 1
    for t in range(2)
    for s in (list(range(1, RX_SECTOR_TRACK, 2)) + list(range(2, RX_SECTOR_TRACK + 1, 2)))
]

# Interleave tables for RX02 in 12-bit mode
RX02_INTERLEAVE_12B = [
    (0 * RX_SECTOR_TRACK) + s - 1
    for s in (
        list(range(1, RX_SECTOR_TRACK + 1, 3))
        + list(range(2, RX_SECTOR_TRACK + 1, 3))
        + list(range(3, RX_SECTOR_TRACK + 1, 3))
    )
]


def get_sector_size(device_size: int) -> int:
    if device_size == RX01_SIZE:
        return RX01_SECTOR_SIZE
    elif device_size == RX02_SIZE:
        return RX02_SECTOR_SIZE
    else:
        return BLOCK_SIZE


def rxfactr(blkno: int, sector_size: int) -> int:
    """
    Calculates the physical position on the disk for a given logical sector
    """
    if sector_size == RX01_SECTOR_SIZE or sector_size == RX02_SECTOR_SIZE:
        track = blkno // RX_SECTOR_TRACK + 1
        i = (blkno % RX_SECTOR_TRACK) << 1
        if i >= RX_SECTOR_TRACK:
            i += 1
        sector = ((i + (6 * (track - 1))) % RX_SECTOR_TRACK) + 1
        if track >= RX_TRACK_DISK:
            track = 0
        position = track * 3328 + (sector - 1) * sector_size
    else:
        position = blkno * BLOCK_SIZE
    return position


def rxfactr_12bit(block_number: int, sector_size: int) -> t.List[int]:
    """
    Calculates the physical position on the disk for a given logical sector
    for RX01 and RX02 in 12-bit mode
    """
    if sector_size == RX01_SECTOR_SIZE:
        interleave = RX01_INTERLEAVE_12B
    else:
        interleave = RX02_INTERLEAVE_12B
    sectors_per_block = BLOCK_SIZE // sector_size  # 4 for RX01, 2 for RX02
    repeat = len(interleave) // sectors_per_block
    base = (block_number // repeat) * repeat * BLOCK_SIZE // RX01_SECTOR_SIZE
    offset = block_number % repeat
    skip = RX_SECTOR_TRACK * sector_size
    result = []
    # Read the sectors
    for i in range(BLOCK_SIZE // sector_size):
        sector = base + interleave[(offset * sectors_per_block) + i]
        position = sector * sector_size + skip
        result.append(position)
    return result


def rx_extract_12bit_words(byte_data: bytes, position: int, sector_size: int) -> t.List[int]:
    """
    Extracts 64 12-bit words from the first 96-bytes of a 128 bytes array
    or the first 128 12-bit words from the first 192-bytes of a 256 bytes array.

    RX01 has 64 12 bit words per sector.
    The words are bit packed into the first 96 bytes.
    RX02 has 128 12 bit words per sector.
    The words are bit packed into the first 192 bytes.
    """
    if sector_size == RX01_SECTOR_SIZE:
        byte_array = byte_data[position : position + 96]
    elif sector_size == RX02_SECTOR_SIZE:
        byte_array = byte_data[position : position + 192]
    else:
        raise ValueError(f"Invalid sector size: {sector_size}")

    bit_buffer = int.from_bytes(byte_array, byteorder="big")
    total_bits = len(byte_array) * 8  # Total number of bits
    total_words = total_bits // 12  # Total number of 12-bit words
    assert total_words == 64 or total_words == 128

    words = []
    for i in range(total_words):
        # Extract the 12 bits corresponding to the current word
        word = (bit_buffer >> (total_bits - 12 * (i + 1))) & 0xFFF  # Mask the last 12 bits
        words.append(word)

    return words


def rx_pack_12bit_words(words: t.List[int], position: int, sector_size: int) -> bytes:
    """
    Converts a list of 12-bit words back to a byte array for RX01 or RX02.

    RX01: Converts 64 12-bit words to a 96-byte array.
    RX02: Converts 128 12-bit words to a 192-byte array.
    """
    if sector_size == RX01_SECTOR_SIZE:
        expected_words = 64
    elif sector_size == RX02_SECTOR_SIZE:
        expected_words = 128
    else:
        raise ValueError(f"Invalid sector size: {sector_size}")

    words = words[position : position + expected_words]
    if len(words) != expected_words:
        raise ValueError(f"Expected {expected_words} words, but got {len(words)}")

    bit_buffer = 0
    for word in words:
        bit_buffer = (bit_buffer << 12) | (word & 0xFFF)  # Pack each 12-bit word

    byte_length = (len(words) * 12 + 7) // 8  # Calculate number of bytes needed
    return bit_buffer.to_bytes(byte_length, byteorder="big")


class RXBlockDevice(BlockDevice):
    """
    Block device with RX01/RX02 support.
    """

    is_rx: bool  # True if this device is a RX01/RX02

    def __init__(self, file: "AbstractFile"):
        super().__init__(file)
        self.sector_size = get_sector_size(self.size)
        self.is_rx = self.sector_size in (RX01_SECTOR_SIZE, RX02_SECTOR_SIZE)

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        if self.is_rx:
            if number_of_blocks == READ_FILE_FULL:
                number_of_blocks = int(math.ceil(self.size / BLOCK_SIZE))
            if block_number < 0 or number_of_blocks < 0:
                raise OSError(errno.EIO, os.strerror(errno.EIO))
            ret = []
            start_sector = block_number * BLOCK_SIZE // self.sector_size
            for i in range(0, number_of_blocks * BLOCK_SIZE // self.sector_size):
                blkno = start_sector + i
                position = rxfactr(blkno, self.sector_size)
                self.f.seek(position)  # not thread safe...
                ret.append(self.f.read(self.sector_size))
            return b"".join(ret)
        else:
            return super().read_block(block_number, number_of_blocks)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        if block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if self.is_rx:
            start_sector = block_number * BLOCK_SIZE // self.sector_size
            for i in range(0, number_of_blocks * BLOCK_SIZE // self.sector_size):
                blkno = start_sector + i
                position = rxfactr(blkno, self.sector_size)
                self.f.seek(position)  # not thread safe...
                self.f.write(buffer[i * self.sector_size : (i + 1) * self.sector_size])
        else:
            super().write_block(buffer, block_number, number_of_blocks)
